/****************************************************
 * File: gui.cpp
 * Author: Keith Schwarz (htiek@cs.stanford.edu)
 *
 * A program that allows the user to choose a color and
 * learn what the computer thinks the color is named.
 * Because the data is a bit noisy, you may encounter
 * some amusing results when running this program.
 */

#include <iostream>
#include <fstream>
#include <vector>
#include <sstream>
#include <string>
#include <algorithm>
#include <map>
#include <cctype>
#include <iterator>
#include <windows.h>
#include <tchar.h>
#include <commctrl.h>
#include "../KDTree.h"
using namespace std;

/* TCHAR string type.  Windows applications use a special type called TCHAR
 * for all characters that allows the code to compile both on ASCII machines
 * and on multibyte character set machines (i.e. Unicode).  This code builds
 * a string type called tstring using the standard library's basic_string
 * template.  Interestingly, basic_string is what's used to build the standard
 * string type; string is actually a typedef for basic_string<char>.
 */
typedef basic_string<TCHAR> tstring;

/* Message used by the loading thread to tell the GUI what it's currently doing. 
 * This keeps the system responsive even when all the data has not yet been loaded.
 */
static const UINT kNotifyMessage = WM_APP + 1;

/* Possible states that the program can be in while loading data. */
enum { Loading, Done, Failure };

/* Global variables are really awful, but in this situation I was unable to figure
 * out a better way to communicate data between functions.  Consequently, we're stuck
 * with this unfortunate state of affairs.
 */
KDTree<3, string> gKDTree;
HWND gStatusWindow = NULL;
bool gIsReady = false;

/* Conversion functions to convert between C++ strings and TCHAR strings. */
static tstring WidenString(string str) {
	tstring result;
	for (size_t i = 0; i < str.length(); ++i)
		result += str[i];
	return result;
}
static string NarrowString(tstring str) {
	string result;
	for (size_t i = 0; i < str.length(); ++i)
		result += char(str[i]);
	return result;
}

/* This function is part of an inelegant hack to get the window to display as
 * it does.  We take the system color chooser, then delete all of the buttons
 * out of it other than the "OK" button.  That button is then hijacked to
 * name the color rather than close the window.  This function is a callback
 * that detects whether a window is a button other than "OK" and destroys it
 * if it is.
 */
static BOOL CALLBACK RemoveAllButtonsExcept(HWND curr, LPARAM ok) {
	/* Skip the "OK" button */
	if (curr == reinterpret_cast<HWND>(ok))
		return TRUE;

	/* Get the name of the window class (the type of window this
	 * window is.)
	 */
	TCHAR buffer[1024] = {0};
	GetClassName(curr, buffer, 1024);

	/* Convert it to a lowercase string. */
	string actualName = NarrowString(buffer);
	transform(actualName.begin(), actualName.end(), actualName.begin(), ::tolower);

	/* If this window is a button, remove it. */
	if (actualName == "button")
		DestroyWindow(curr);

	return TRUE;
}

/* Loads the color name data from disk. */
static bool LoadDataSet(HWND hWnd) {
	/* Open the file; fail if we can't. */
	ifstream input("colors.txt", ios::binary);
	if (!input) return false;

	/* Read how many entries there are and preallocate space for them. */
	size_t count;
	if (!(input >> count)) return false;

	/* Skip the newline. */
	if (!input.ignore(1))
		return false;

	/* Allocate these out here for efficiency. */
	string line;
	size_t read = 0;

	/* Keep reading data out of the file and parsing it to color data. */
	while (true) {
		/* Read RGB */
		char colorBuffer[3];
		input.read(colorBuffer, streamsize(3));

		/* Read name size. */
		char bufferSize = char(input.get());
		size_t toRead = (unsigned char)(bufferSize);

		/* Read name from file.  Note that because characters range from
		 * 0 - 255, we can't overrun this buffer.
		 */
		char nameBuffer[256];
		input.read(nameBuffer, streamsize(toRead));

		/* Check for stream integrity. */
		if (!input) 
			break;

		/* Add to the data set. */
		Point<3> currPt;
		currPt[0] = (unsigned char)(colorBuffer[0]);
		currPt[1] = (unsigned char)(colorBuffer[1]);
		currPt[2] = (unsigned char)(colorBuffer[2]);
		gKDTree.insert(currPt, string(nameBuffer, nameBuffer + toRead));

		/* Keep the GUI informed of what's going on. */
		if (++read % 10000 == 0)
			PostMessage(hWnd, kNotifyMessage, read, Loading);
	}

	/* Ensure we read enough. */
	return read == count;
}

/* Thread function to load in color data and build a kd-tree out of it. */
static DWORD WINAPI InitializeKDTree(LPVOID data) {
	/* Recover the handle of the main window. */
	HWND hWnd = static_cast<HWND>(data);

	/* Tell the window that we're loading data. */
	PostMessage(hWnd, kNotifyMessage, 0, Loading);
	
	/* Load in the color data from the file. */
	if (!LoadDataSet(hWnd)) {
		PostMessage(hWnd, kNotifyMessage, 0, Failure);
		return 0;
	}

	PostMessage(hWnd, kNotifyMessage, 0, Done);
	gIsReady = true;
	return 1;
}

/* Initializes the thread that loads all of the external data. */
static bool FireLoadThread(HWND hDlg) {
	return CreateThread(NULL, 0, InitializeKDTree, hDlg, 0, NULL) != NULL;
}

/* Given a handle to the Choose Color dialog, alters the window into the main
 * window for this program.
 */
static bool InitializeDialog(HWND hDlg) {
	/* Change the name of the window so that it looks more like a standalone
	 * application.
	 */
	SetWindowText(hDlg, _T("Color Lookup"));

	/* Find the OK button and change it to read "Name Color" */
	HWND ok = GetDlgItem(hDlg, IDOK);
	if (ok == NULL)
		return false;

	if (!SetWindowText(ok, _T("Name Color")))
		return false;

	/* Delete all other buttons from the window. */
	if (!EnumChildWindows(hDlg, RemoveAllButtonsExcept, reinterpret_cast<LPARAM>(ok)))
		return false;

	/* Increase the size of the window to accommodate the status line we'll be adding. */
	RECT rc;
	if(!GetWindowRect(hDlg, &rc))
		return false;
	rc.bottom += 30; // Total hack; might not be enough!

	/* Compute the dimensions of the window from the bounding rect. */
	int width  = rc.right - rc.left;
	int height = rc.bottom - rc.top;

	/* Compute where the window should go in order to be centered. */
	int x = (GetSystemMetrics(SM_CXSCREEN) - width)  / 2;
	int y = (GetSystemMetrics(SM_CYSCREEN) - height) / 2;

	/* Move the window there. */
	if(!SetWindowPos(hDlg, HWND_TOP, x, y, width, height, SWP_SHOWWINDOW))
		return false;

	/* Create a status line for output. */
	gStatusWindow = CreateWindowEx(0, STATUSCLASSNAME, _T("Loading..."),
		                    WS_CHILD | WS_VISIBLE, 
							0, 0, 0, 0,
							hDlg, NULL, GetModuleHandle(NULL), NULL);
	if (gStatusWindow == NULL)
		return false;

	/* Set the window icon to something pretty. */
	SendMessage(hDlg, WM_SETICON, ICON_BIG,   reinterpret_cast<LPARAM>(LoadIcon(GetModuleHandle(NULL), _T("icon"))));
	SendMessage(hDlg, WM_SETICON, ICON_SMALL, reinterpret_cast<LPARAM>(LoadIcon(GetModuleHandle(NULL), _T("icon"))));

	/* Start loading data. */
	if (!FireLoadThread(hDlg))
		return false;

	return true;
}

/* Given the name of a color, uses 3-NN lookup to get the computer's best guess of the color name. */
static void ReportColorName(COLORREF color) {
	/* Ensure the program has loaded and is ready to go. */
	if (!gIsReady || !gStatusWindow)
		return;

	/* Convert from COLORREF into a vector<double> */
	Point<3> colorV;
	colorV[0] = GetRValue(color);
	colorV[1] = GetGValue(color);
	colorV[2] = GetBValue(color);

	/* Perform the lookup and get the color. */
	string closest = "Color name: " + gKDTree.kNNValue(colorV, 3);

	/* Display the result. */
	SetWindowText(gStatusWindow, WidenString(closest).c_str());
}

/* Responds to a notification from the loading thread. */
static void HandleNotification(HWND hDlg, WPARAM wParam, LPARAM lParam) {
	/* Based on the message type, react differently. */
	switch (lParam) {
	/* Report how many items we've loaded. */
	case Loading:
		{
			stringstream converter;
			converter << "Loaded " << wParam << " data points...";
			SetWindowText(gStatusWindow, WidenString(converter.str()).c_str());
			break;
		}

	/* Report that we're ready to go. */
	case Done:
		SetWindowText(gStatusWindow, _T("Ready"));
		break;

	/* Report an error and quit the program. */
	case Failure:
		PostMessage(hDlg, WM_COMMAND, IDABORT, 0);
		MessageBox(hDlg, _T("A fatal error occurred loading data.  This program will now exit"), _T("Fatal Error"), MB_ICONSTOP | MB_OK);
		break;
	}
}

/* A callback function that hooks into the color chooser's normal routine and
 * uses the color chooser to name colors rather than returning them.  Most of
 * this is evil hackery to convert the dialog into something it really isn't.
 */
static UINT_PTR CALLBACK HookColorFunction(HWND hDlg, UINT uiMsg, WPARAM wParam, LPARAM lParam) {
	/* On initialization, reshape the dialog to its final form. */
	if (uiMsg == WM_INITDIALOG) {
		if (!InitializeDialog(hDlg))
			PostMessage(hDlg, WM_COMMAND, IDABORT, 0);
	}

	/* If the user clicks the "Name Color" button, obtain the name of the color
	 * and display it to the user.
	 */
	if (uiMsg == RegisterWindowMessage(COLOROKSTRING)) {
		ReportColorName(reinterpret_cast<LPCHOOSECOLOR>(lParam)->rgbResult);	
		return 1; // Don't have the window do normal processing; we want it open.
	}

	/* If there is any news from the loading thread, report what it is. */
	if (uiMsg == kNotifyMessage)
		HandleNotification(hDlg, wParam, lParam);

	return 0;
}

/* Generates a random color. */
COLORREF RandomColor() {
	return RGB(rand() % 256, rand() % 256, rand() % 256);
}

/* Main program entry point.  Creates the window and turns control over
 * to the event loop.
 */
int WINAPI WinMain(HINSTANCE, HINSTANCE, LPSTR, int) {
	InitCommonControls();

	/* Create a parameters object storing information about how we want the
	 * color dialog to be.
	 */
	CHOOSECOLOR c;
	memset(&c, 0, sizeof(c));
	c.lStructSize = sizeof(CHOOSECOLOR);

	/* Generate space to hold the sixteen custom colors and fill them
	 * with random values.
	 */
	COLORREF storedValues[16];
	generate(storedValues, storedValues + 16, RandomColor);
	c.lpCustColors = storedValues;

	/* Hook our custom callback into the event handler; this makes the dialog
	 * act entirely different from what's expected.
	 */
	c.lpfnHook = HookColorFunction;

	/* Pick a nice gray as the starting color. */
	c.rgbResult = RGB(192, 192, 192);
	c.Flags = CC_ANYCOLOR | CC_FULLOPEN | CC_ENABLEHOOK | CC_RGBINIT;

	/* Fire off the color chooser; program will terminate when it closes. */
	ChooseColor(&c);
	return 0;
}
