/*********************************************
 * File: gui.cpp
 * Author: Keith Schwarz (htiek@cs.stanford.edu)
 *
 * Implementation of the Digit Classifier program.  This
 * program presents the user a canvas on which to draw
 * a digit.  When the user clicks the "Go" button, the image is
 * centered and normalized to a predefined height using a
 * KD tree, then classified using the k-NN algorithm.
 */

#include <iostream>
#include <string>
#include <fstream>
#include <vector>
#include <sstream>
#include <queue>
#include <windows.h>
#include <tchar.h>
#include <commctrl.h>
#include <limits>
#include "../KDTree.h"
#include "grid.h"
using namespace std;

/* One of the Windows headers defines max using #define.  This
 * conflicts with the STL algorithms and messes up other parts
 * of the code.  We #undef it here to avoid these issues.
 */
#undef max

/* This message is used by the loading thread to notify the
 * main window about what it's doing while loading.
 */
static const UINT kNotifyMessage   = WM_APP + 1;
enum { Loading, Done, Failure };


/* TCHAR string type.  Windows applications use a special type called TCHAR
 * for all characters that allows the code to compile both on ASCII machines
 * and on multibyte character set machines (i.e. Unicode).  This code builds
 * a string type called tstring using the standard library's basic_string
 * template.  Interestingly, basic_string is what's used to build the standard
 * string type; string is actually a typedef for basic_string<char>.
 */
typedef basic_string<TCHAR> tstring;

/* Name of the main window classes.  All windows have a string identifying what "class"
 * they belong to.  Much in the same way that C++ objects are instances of a particular
 * C++ class, Windows windows are instances of a particular window class.  This is the
 * name we give to the main windows.
 */
static const TCHAR * const kCanvasClassName = _T("DigitClassifierCanvasClass");
static const TCHAR * const kWindowClassName = _T("DigitClassifierMainWindowClass");

/* Constant: kImageDimension
 * Value: The size of one side of an image.
 */
const size_t kImageDimension = 28;

/* Constant: kImageSize
 * Value: The number of pixels in an image.
 */
const size_t kImageSize = kImageDimension * kImageDimension;

/* Dimensions of the image (and by proxy, the window) */
static const int kPixelsPerImageElem = 7;

/* Size of the canvas and main window. */
static const int  kCanvasWidth     = kImageDimension * kPixelsPerImageElem;
static const int  kCanvasHeight    = kImageDimension * kPixelsPerImageElem;

static const int  kWindowWidth     = 3 * kCanvasWidth / 2;
static const int  kWindowHeight    = 2 * kCanvasHeight;

/* This information is necessary to perform all of the 
 * digit analysis.  It will be stored as auxiliary window
 * data in the main window, hence the name.
 */
struct WindowAuxiliaryData {
	HWND       hWnd;    // Main window
	HWND       hStatus; // Status window
	HWND       hCanvas; // Canvas window
	grid<bool> canvasContents; // Drawn image
	bool       isReady; // Whether the KDTree is ready

	KDTree<kImageSize, unsigned char>    kd;  // KD tree for lookups
	queue< Point<kImageSize> > analysisQueue; // List of grids to analyze

	HANDLE     canvasSemaphore; // Binary semaphore used to guard the canvas.
	HANDLE     queueSemaphore;  // Binary semaphore used to guard the task queue.
	HANDLE     queueReady;      // n-ary  semaphore used to pause processing task.
};

/* Helper class to auto acquire/release a semaphore.  This class
 * ensures that a lock is never taken without being released.
 */
class AutoUnlock {
public:
	/* Constructor acquires a lock. */
	explicit AutoUnlock(HANDLE s) : semaphore(s) {
		/* Acquire the semaphore and fail if something bad happens with it. */
		if (WaitForSingleObject(semaphore, INFINITE) != WAIT_OBJECT_0)
			exit(-1);
	}

	/* Destructor releases it. */
	~AutoUnlock() {
		if (!ReleaseSemaphore(semaphore, 1, NULL))
			exit(-1);
	}

private:
	HANDLE semaphore;

	AutoUnlock(const AutoUnlock&); // No copying
	AutoUnlock& operator= (const AutoUnlock&); // No assigning
};

/* Helper function that, given a window, returns the auxiliary data
 * associated with it.  There's a nasty typecast involved; hence the
 * wrapper function.
 */
WindowAuxiliaryData* GetAuxData(HWND hWnd) {
	return reinterpret_cast<WindowAuxiliaryData*>(GetWindowLongPtr(hWnd, GWLP_USERDATA));
}

/* Loads all of the image examples from disk. */
static bool LoadData(HWND hWnd, KDTree<kImageSize, unsigned char>& kd) try {
	/* These next lines tell the stream to generate exceptions on failure.
	 * This greatly simplifies processing.
	 */
	ifstream images; // Image file
	images.exceptions(ios::failbit | ios::badbit);
	images.open("training-images", ios::binary);

	ifstream labels; // Labels for those images
	labels.exceptions(ios::failbit | ios::badbit);
	labels.open("training-labels", ios::binary);

	/* Each file begins with a "magic number" identifying the file.
	 * These numbers are stored as raw bytes in low-endian form.
	 * These next few lines read the data from the file and confirm
	 * that the values are what we'd expect them to be.
	 */
	char buffer[4];
	images.read(buffer, 4);

	/* memcmp is a function which compares raw memory and returns an integer
	 * based on their lexicographical ordering.  In particular, it should
	 * return zero if the values match.
	 */
	if (memcmp(buffer, "\x00\x00\x08\x03", 4) != 0)
		return false;

	/* Now read how many images there are.  This also requires some memory hackery
	 * because the integer is stored in low-endian format and we need to reconstruct
	 * it.
	 */
	images.read(buffer, 4);
	size_t numImages = ((unsigned char)buffer[0] << 24) +
		               ((unsigned char)buffer[1] << 16) + 
					   ((unsigned char)buffer[2] << 8) + 
					    (unsigned char)buffer[3];
	
	/* Next in the stream are the image dimensions.  We really don't care about these, so we just
	 * skip them.
	 */
	images.ignore(streamsize(8));

	/* Now repeat this process for the labels file. */
	labels.read(buffer, 4);
	if (memcmp(buffer, "\x00\x00\x08\x01", 4) != 0)
		return false;

	labels.read(buffer, 4);
	size_t numLabels = ((unsigned char)buffer[0] << 24) + 
		               ((unsigned char)buffer[1] << 16) + 
					   ((unsigned char)buffer[2] << 8) + 
					    (unsigned char)buffer[3];

	/* Confirm that we have the same number of images as labels. */
	if (numLabels != numImages) return false;

	/* Read all images. */
	for (size_t i = 0; i < numLabels; ++i) {
		/* Images are stored as byte arrays ranging from 0 - 255.  We'll
		 * read the whole image in and then convert it to binary white/black
		 * to match our input format.
		 */
		char buffer[kImageSize];
		images.read(buffer, kImageSize);

		Point<kImageSize> pt;

		/* Convert from 0 - 255 to -1 - +1 */
		for (size_t j = 0; j < kImageSize; ++j) {
			unsigned char value = buffer[j];
			pt[j] = (value > 0 ? 1.0 : -1.0);
		}

		/* Read label */
		char byte;
		labels.get(byte);

		/* Add this image! */
		kd.insert(pt, byte + '0');

		/* Keep the GUI informed about what's going on. */
		if (i % 1000 == 0)
			SendMessage(hWnd, kNotifyMessage, i, Loading);
	}
	return true;
} catch (const exception&) {
	/* On error, signal failure. */
	return false;
}

/* Helper function to convert a regular string into a tstring. */
static tstring WidenString(string str) {
	tstring result;
	for (size_t i = 0; i < str.length(); ++i)
		result += str[i];
	return result;
}

/* Helper function used as an algorithms callback.  It converts a boolean
 * true/false into a real-valued 1/-1
 */
static double BoolToValue(bool b) {
	return b ? 1.0 : -1.0;
}

/* Redraws the user's image in response to window invalidation. */
static void RedrawCanvas(HWND hWnd, WindowAuxiliaryData* wad) {
	PAINTSTRUCT ps;
	HDC dc = BeginPaint(hWnd, &ps);

	{
		/* Lock the canvas so no one else can use it. */
		AutoUnlock au(wad->canvasSemaphore);
			
		/* Traverse the image and display it pixel-by-pixel. */
		for (int row = 0; row < int(wad->canvasContents.numRows()); ++row) {
			for (int col = 0; col < int(wad->canvasContents.numCols()); ++col) {
				/* Each pixel is actually represented as a block on-screen.  This
				 * code computes the block and displays it.
				 */
				RECT rc;
				SetRect(&rc, col * kPixelsPerImageElem,       row * kPixelsPerImageElem,
							(col + 1) * kPixelsPerImageElem, (row + 1) * kPixelsPerImageElem);
				FillRect(dc, &rc, (HBRUSH)GetStockObject(wad->canvasContents[row][col] ? BLACK_BRUSH : WHITE_BRUSH));
			}
		}
	}

	/* Cleanup. */
	ReleaseDC(hWnd, dc);
	EndPaint(hWnd, &ps);
}

/* Updates the canvas after the client draws on it. */
static void DrawOnCanvas(HWND hWnd, int x, int y, WindowAuxiliaryData* wad) {
	x /= kPixelsPerImageElem;
	y /= kPixelsPerImageElem;

	/* Lock the canvas so that we don't get any race conditions. */
	{
		AutoUnlock au(wad->canvasSemaphore);

		/* Draw in a 3x3 box centered on the click location. */
		for (int xP = x - 1; xP <= x + 1; ++xP) {
			for (int yP = y - 1; yP <= y + 1; ++yP) {
				if (xP >= 0 && yP >= 0 && xP < int(wad->canvasContents.numCols()) && yP < int(wad->canvasContents.numRows()))
					wad->canvasContents[yP][xP] = true;
			}
		}
	}
	
	/* Tell the window that it needs to redraw this part. */
	RECT rc;
	SetRect(&rc, (x - 1) * kPixelsPerImageElem, (y - 1) * kPixelsPerImageElem,
		         (x + 2) * kPixelsPerImageElem, (y + 2) * kPixelsPerImageElem);
	InvalidateRect(hWnd, &rc, FALSE);
	UpdateWindow(hWnd);
}

/* Canvas window's event handler.  This function is implicitly called by the system whenever
 * an event occurs.  We use this to dispatch commands to the proper handlers.
 */
static LRESULT CALLBACK HandleCanvasMessage (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	/* Ensure we have aux data and gracefully fail if we don't. */
	WindowAuxiliaryData* wad = GetAuxData(hWnd);
	if (!wad) return DefWindowProc(hWnd, message, wParam, lParam);

	/* To paint the canvas, we need to redraw the image. */
	if (message == WM_PAINT) {
		RedrawCanvas(hWnd, wad);
		return 0;
	}

	/* On a click, draw on the canvas. */
	else if (message == WM_LBUTTONDOWN || (message == WM_MOUSEMOVE && (wParam & MK_LBUTTON))) {
		DrawOnCanvas(hWnd, LOWORD(lParam), HIWORD(lParam), wad);
		return 0;
	}
	
	/* Otherwise just do whatever the system default is. */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/* Responds to a notification from the loading thread. */
static void HandleNotification(HWND hWnd, HWND hStatus, WPARAM wParam, LPARAM lParam) {
	/* Convert bad arguments into failure. */
	if (!hStatus) lParam = Failure;

	/* Based on the message type, react differently. */
	switch (lParam) {
	/* Report how many items we've loaded. */
	case Loading:
		{
			stringstream converter;
			converter << "Loaded " << wParam << " data points...";
			SetWindowText(hStatus, WidenString(converter.str()).c_str());
			break;
		}

	/* Report that we're ready to go. */
	case Done:
		SetWindowText(hStatus, _T("Ready"));
		break;

	/* Report an error and quit the program. */
	case Failure:
		DestroyWindow(hWnd);
		MessageBox(NULL, _T("A fatal error occurred loading data.  This program will now exit"), _T("Fatal Error"), MB_ICONSTOP | MB_OK);
		break;
	}
}

/* Given an image, shifts it over so that it hugs the top and left edges
 * of the image.
 */
static void TranslateToOrigin(grid<bool>& image) {
	size_t lowX = numeric_limits<size_t>::max(), lowY = numeric_limits<size_t>::max();

	/* Find the lowest x and y value. */
	for (size_t x = 0; x < image.numCols(); ++x) {
		for (size_t y = 0; y < image.numRows(); ++y) {
			if (image[y][x]) {
				lowX = min(x, lowX);
				lowY = min(y, lowY);
			}
		}
	}

	/* Rebuild the grid. */
	grid<bool> result(image.numRows(), image.numCols());
	for (size_t x = 0; x < image.numCols(); ++x)
		for (size_t y = 0; y < image.numRows(); ++y)
			if (image[y][x])
				result[y - lowY][x - lowX] = true;

	image = result;
}

/* Given an image hugging the top-left corner, shifts it over so that it's
 * centered.
 */
static void CenterImage(grid<bool>& image) {
	/* Find the max x and y coordinates. */
	size_t highX = 0, highY = 0;

	/* Find the maximum x and y value. */
	for (size_t x = 0; x < image.numCols(); ++x) {
		for (size_t y = 0; y < image.numRows(); ++y) {
			if (image[y][x]) {
				highX = max(x, highX);
				highY = max(y, highY);
			}
		}
	}

	/* We now have the dimensions.  If either are zero, we're done. */
	if (!highX || !highY) return;

	/* Figure out how much more wiggle room we have and shift it over by
	 * that amount.
	 */
	size_t dx = (image.numCols() - highX) / 2;
	size_t dy = (image.numRows() - highY) / 2;

	/* Rebuild the grid. */
	grid<bool> result(image.numRows(), image.numCols());
	for (size_t x = 0; x < image.numCols(); ++x)
		for (size_t y = 0; y < image.numRows(); ++y)
			if (image[y][x])
				result[y + dy][x + dx] = true;

	image = result;
}

/* Given an image whose upper corner is the origin, rescales it to fit in a 20 x 20 box. */
static void RescaleImage(grid<bool>& image) {
	/* Find the max x and y coordinates. */
	size_t highX = 0, highY = 0;

	/* Find the maximum x and y value. */
	for (size_t x = 0; x < image.numCols(); ++x) {
		for (size_t y = 0; y < image.numRows(); ++y) {
			if (image[y][x]) {
				highX = max(x, highX);
				highY = max(y, highY);
			}
		}
	}

	/* We now have the dimensions.  If either are zero, we're done. */
	if (!highX || !highY) return;

	/* Compute the new width and height based on the extents. */
	size_t width, height;
	if (highX > highY) {
		width = 20;
		height = size_t(ceil((20.0 * highY) / highX));
	} else {
		height = 20;
		width = size_t(ceil((20.0 * highX) / highY));
	}

	/* Use a nearest-neighbor scaling algorithm.  This works as follows.  First,
	 * we add a data point for each pixel that's normalized to [0, 1] based on
	 * its x, y position.  Next, we iterate over the normalized size and set each
	 * pixel to its nearest-neighbor value.
	 */
	KDTree<2, bool> nnGrid;
	for (size_t x = 0; x < highX; ++x) {
		for (size_t y = 0; y < highY; ++y) {
			Point<2> pt;
			pt[0] = x;
			pt[1] = y;
			nnGrid.insert(pt, image[y][x]);
		}
	}

	/* Use nearest-neighbor lookup to interpolate the points. */
	grid<bool> result(image.numRows(), image.numCols());
	for (size_t x = 0; x < width; ++x) {
		for (size_t y = 0; y < height; ++y) {
			Point<2> pt;
			pt[0] = x;
			pt[1] = y;
			result[y][x] = nnGrid.kNNValue(pt, 1);
		}
	}

	image = result;
}

static void ResetCanvas(WindowAuxiliaryData* wad) {
	/* Lock the canvas to avoid race conditions, then clear it. */
	{
		AutoUnlock au(wad->canvasSemaphore);
		fill(wad->canvasContents.begin(), wad->canvasContents.end(), false);
	}

	/* Force a redraw. */
	RECT rc;
	SetRect(&rc, 0, 0, kCanvasWidth, kCanvasHeight);
	InvalidateRect(wad->hCanvas, &rc, FALSE);
	UpdateWindow(wad->hCanvas);
}

/* Initiates a classification request. */
static void InitiateRequest(HWND hWnd, WindowAuxiliaryData* wad) {
	/* Extract contents from the canvas. */
	grid<bool> image;
	{
		AutoUnlock au(wad->canvasSemaphore);
		image = wad->canvasContents;
	}

	/* Perform preprocessing on image to scale and center it. */
	TranslateToOrigin(image);
	RescaleImage(image);
	CenterImage(image);

	/* Convert from grid to vector. */
	Point<kImageSize> dataPoint;
	transform(image.begin(), image.end(), dataPoint.begin(), BoolToValue);

	/* Add to processing queue. */
	{
		AutoUnlock au(wad->queueSemaphore);
		wad->analysisQueue.push(dataPoint);
	}

	/* Wake up the processing thread. */
	if (!ReleaseSemaphore(wad->queueReady, 1, NULL))
		SendMessage(hWnd, kNotifyMessage, 0, Failure);
}

/* Handles a child command message from the main window. */
static void ProcessCommand(HWND hWnd, WindowAuxiliaryData* wad, WPARAM wParam) {
	/* Initiate a request */
	if (LOWORD(wParam) == IDOK)
		InitiateRequest(hWnd, wad);

	/* Clear the canvas. */
	if (LOWORD(wParam) == IDABORT)
		ResetCanvas(wad);
}

/* Main window's event handler.  This function is implicitly called by the system whenever
 * an event occurs.  We use this to dispatch commands to the proper handlers.
 */
static LRESULT CALLBACK HandleMessage (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	/* If the window closes, quit the program. */
	if (message == WM_DESTROY) {
		PostQuitMessage(0);
		return 0;
	}
	
	/* Check for aux data and fall back on default if unavailable. */
	WindowAuxiliaryData* wad = GetAuxData(hWnd);
	if (!wad) return DefWindowProc(hWnd, message, wParam, lParam);

	/* On a status notification, update the status window. */
	if (message == kNotifyMessage)
		HandleNotification(hWnd, wad->hStatus, wParam, lParam);

	/* On a child command, divert processing to the proper routine. */
	if (message == WM_COMMAND)
		ProcessCommand(hWnd, wad, wParam);
	
	/* Otherwise just do whatever the system default is. */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/* Configures the window class for the drawing canvas. */
static bool InitializeCanvasWindowClass(HINSTANCE hInst) {
	WNDCLASSEX windowClass;
	windowClass.cbSize = sizeof(WNDCLASSEX);
	windowClass.style = CS_OWNDC | CS_DBLCLKS;
	windowClass.lpfnWndProc = HandleCanvasMessage;
	windowClass.cbWndExtra = 0;
	windowClass.cbClsExtra = 0;
	windowClass.hInstance = hInst;

	windowClass.hIcon = LoadIcon(hInst, _T("icon"));
	windowClass.hCursor = LoadCursor(NULL, IDC_CROSS);
	windowClass.hbrBackground = (HBRUSH)COLOR_APPWORKSPACE;
	windowClass.lpszMenuName = NULL;
	windowClass.lpszClassName = kCanvasClassName;
	windowClass.hIconSm = LoadIcon(hInst, _T("icon"));
	
	/* Make sure registration worked! */
	return RegisterClassEx(&windowClass) != 0;
}

/* Configures the window class for the main window.  This is all fairly technical
 * and is completely uninteresting.
 */
static bool InitializeMainWindowClass(HINSTANCE hInst) {
	WNDCLASSEX windowClass;
	windowClass.cbSize = sizeof(WNDCLASSEX);
	windowClass.style = CS_OWNDC | CS_DBLCLKS;
	windowClass.lpfnWndProc = HandleMessage;
	windowClass.cbWndExtra = 0;
	windowClass.cbClsExtra = 0;
	windowClass.hInstance = hInst;

	windowClass.hIcon = LoadIcon(hInst, _T("icon"));
	windowClass.hCursor = LoadCursor(NULL, IDC_CROSS);
	windowClass.hbrBackground = (HBRUSH)COLOR_APPWORKSPACE;
	windowClass.lpszMenuName = NULL;
	windowClass.lpszClassName = kWindowClassName;
	windowClass.hIconSm = LoadIcon(hInst, _T("icon"));
	
	/* Make sure registration worked! */
	return RegisterClassEx(&windowClass) != 0;
}

/* Configures the main window after creation. */
static bool InitializeWindow(HINSTANCE hInst, HWND hWnd, WindowAuxiliaryData* wad) {
	/* Force the window to be the right size. */
	RECT rv;
	GetWindowRect(hWnd, &rv);
	POINT windowPos = {rv.left, rv.top};

	/* Get a rectangle holding the new coordinates */
	RECT rc;
	SetRect(&rc, 0, 0, kWindowWidth, kWindowHeight);
	AdjustWindowRectEx(&rc, WS_BORDER | WS_SYSMENU, FALSE, 0);
	
	/* Put the window there */
	MoveWindow(hWnd, rc.left, rc.top, rc.right-rc.left, rc.bottom-rc.top, TRUE);

	/* Get new size */
	RECT newsize;
	GetClientRect(hWnd, &newsize);

	/* Adjust based on new size information */
	rc.right  += kWindowWidth - newsize.right;
	rc.bottom += kWindowHeight - newsize.bottom;

	/* Shift */
	OffsetRect(&rc, windowPos.x, windowPos.y);

	/* Move and show window */
	MoveWindow(hWnd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);

	/* Create the canvas. */
	wad->hCanvas = CreateWindowEx(WS_EX_DLGMODALFRAME, kCanvasClassName, _T(""), WS_CHILD | WS_VISIBLE, (kWindowWidth - kCanvasWidth) / 2, (kWindowHeight / 8), kCanvasWidth, kCanvasHeight, hWnd, NULL, hInst, NULL);
	if (!wad->hCanvas) return false;

	/* Create the status window. */
	wad->hStatus = CreateWindowEx(0, STATUSCLASSNAME, _T("Loading..."), WS_CHILD | WS_VISIBLE, 0, 0, 0, 0, hWnd, NULL, hInst, NULL);
	if (!wad->hStatus) return false;

	/* Create the Go and Clear buttons. */
	if (!CreateWindowEx(0, _T("Button"), _T("Go"), WS_CHILD | WS_VISIBLE, kWindowWidth / 4, (kCanvasHeight * 3) / 2, kWindowWidth / 5, 30, hWnd, reinterpret_cast<HMENU>(IDOK), hInst, NULL))
		return false;
	if (!CreateWindowEx(0, _T("Button"), _T("Clear"), WS_CHILD | WS_VISIBLE, kWindowWidth / 2 + kWindowWidth / 20, (kCanvasHeight * 3) / 2, kWindowWidth / 5, 30, hWnd, reinterpret_cast<HMENU>(IDABORT), hInst, NULL))
		return false;

	/* Update the window aux data so that we can look this up later. */
	SetLastError(0);
	if(!SetWindowLongPtr(hWnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(wad)) && GetLastError())
		return false;

	/* Update the canvas aux data so that we can look this up later. */
	SetLastError(0);
	if(!SetWindowLongPtr(wad->hCanvas, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(wad)) && GetLastError())
		return false;

	return true;
}

/* Initializes the window auxiliary data structure. */
static bool InitializeAuxData(WindowAuxiliaryData& wad) {
	/* Default everything to something sensible. */
	wad.canvasContents.resize(kImageDimension, kImageDimension);
	wad.hWnd    = NULL;
	wad.hCanvas = NULL;
	wad.hStatus = NULL;
	wad.isReady = false;

	/* Create a lock for the grid. */
	SetLastError(0);
	wad.canvasSemaphore = CreateSemaphore(NULL, 1, 1, NULL);
	if (wad.canvasSemaphore == NULL || GetLastError() != 0)
		return false;

	/* Create a lock for the grid. */
	SetLastError(0);
	wad.queueSemaphore = CreateSemaphore(NULL, 1, 1, NULL);
	if (wad.queueSemaphore == NULL || GetLastError() != 0)
		return false;

	/* Create a semaphore for the processing queue. */
	SetLastError(0);
	wad.queueReady = CreateSemaphore(NULL, 0, numeric_limits<LONG>::max(), NULL);
	if (wad.queueReady == NULL || GetLastError() != 0)
		return false;

	return true;
}

/* Thread responsible for loading data from the file. */
static DWORD WINAPI LoadingThread(LPVOID data) {
	WindowAuxiliaryData* wad = static_cast<WindowAuxiliaryData*>(data);

	/* Load the data points. */
	if (!LoadData(wad->hWnd, wad->kd)) {
		SendMessage(wad->hWnd, kNotifyMessage, 0, Failure);
		return 0;
	}

	/* Report success. */
	wad->isReady = true;
	PostMessage(wad->hWnd, kNotifyMessage, 0, Done);
	return 1;
}

/* Fires off the loading thread. */
static bool StartLoadingThread(WindowAuxiliaryData* wad) {
	return CreateThread(NULL, 0, LoadingThread, wad, 0, NULL) != NULL;
}

/* Thread to process data and report the result. */
static DWORD WINAPI ProcessingThread(LPVOID data) {
	WindowAuxiliaryData* wad = static_cast<WindowAuxiliaryData*>(data);

	while (true) {
		/* Wait for someone to give us work to do. */
		if (WaitForSingleObject(wad->queueReady, INFINITE) != WAIT_OBJECT_0) {
			SendMessage(wad->hWnd, kNotifyMessage, 0, Failure);
			return 0;
		}

		/* Grab the next object out of the queue. */
		Point<kImageSize> nextJob;
		{
			AutoUnlock au(wad->queueSemaphore);
			nextJob = wad->analysisQueue.front();
			wad->analysisQueue.pop();
		}

		/* Perform the analysis, but not if the tree doesn't exist. */
		string result;
		if (!wad->isReady) {
			MessageBox(NULL, _T("Not ready"), _T(""), MB_OK);;
			continue;
		}

		/* Indicate that we're processing. */
		SetWindowText(wad->hStatus, _T("Processing..."));
		result += wad->kd.kNNValue(nextJob, 4);

		/* Report the result. */
		SetWindowText(wad->hStatus, WidenString("Computer thinks digit is a " + result).c_str());
	}
}

/* Fires off the processing thread. */
bool StartProcessingThread(WindowAuxiliaryData* wad) {
	return CreateThread(NULL, 0, ProcessingThread, wad, 0, NULL) != NULL;
}

/* Main program entry point.  Set up the window and wait for user input. */
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, int) {
	InitCommonControls();

	/* Create auxiliary structure and initialize it. */
	WindowAuxiliaryData wad;
	if (!InitializeAuxData(wad)) 
		return 0;

	/* Register window classes so we can create main windows and the canvas. */
	if (!InitializeCanvasWindowClass(hInst) || !InitializeMainWindowClass(hInst)) 
		return 0;
	
	/* Create the window */
	wad.hWnd = CreateWindowEx(0, kWindowClassName, _T("Digit Classifier"), WS_VISIBLE | WS_BORDER | WS_SYSMENU, (GetSystemMetrics(SM_CXSCREEN) - kWindowWidth) / 2,(GetSystemMetrics(SM_CYSCREEN) - kWindowHeight) / 2 , kWindowWidth, kWindowHeight, NULL, NULL, hInst, NULL);
	
	/* Setup the window; fail if unable to do so. */
	if(!wad.hWnd || !InitializeWindow(hInst, wad.hWnd, &wad))
		return 0;

	/* Fire off the worker threads. */
	if (!StartLoadingThread(&wad) || !StartProcessingThread(&wad))
		return 0;

	/* Main program loop.  Wait for a message,then go dispatch it. */
	while(true) {
		WaitMessage();

		MSG message;
		while(PeekMessage(&message, NULL, 0, 0, PM_REMOVE)) {			
			/* On exit, kill the program. */
			if(message.message == WM_QUIT)
				return int(message.wParam);
			
			TranslateMessage(&message);
			DispatchMessage(&message);
		}
	}
}
