/*********************************************************************
 * File: gui.cpp
 * Author: Keith Schwarz (htiek@cs.stanford.edu)
 *
 * Implementation of the MapLookup program.  The program displays a map
 * of the world to clients, then gives back information about where
 * various points on the map are.  You should not need to modify this
 * file in your solution.
 */

#include <windows.h>
#include <iostream>
#include <string>
#include <sstream>
#include <cmath>
#include <vector>
#include <numeric>
#include <cassert>
#include <fstream>
#include <map>
#include <tchar.h>
#include <commctrl.h>
#include "../KDTree.h"
using namespace std;

/* TCHAR string type.  Windows applications use a special type called TCHAR
 * for all characters that allows the code to compile both on ASCII machines
 * and on multibyte character set machines (i.e. Unicode).  This code builds
 * a string type called tstring using the standard library's basic_string
 * template.  Interestingly, basic_string is what's used to build the standard
 * string type; string is actually a typedef for basic_string<char>.
 */
typedef basic_string<TCHAR> tstring;

/* Data structure held by the main window to avoid using global variables. */
struct WindowAuxiliaryData {
	KDTree<2, string> kdt;         // Actual lookup tree
	map<string, string> geoLookup; // Map from code names to actual names.
	HWND hStatus;                  // Status window handle (for reporting)
	bool isReady;                  // Whether or not everything has loaded
};

/* Name of the main window class.  All windows have a string identifying what "class"
 * they belong to.  Much in the same way that C++ objects are instances of a particular
 * C++ class, Windows windows are instances of a particular window class.  This is the
 * name we give to the main window.
 */
static const TCHAR * const kWindowClassName = _T("MapLookupMainWindowClass");

/* Dimensions of the image (and by proxy, the window) */
static const int  kWindowWidth     = 1024;
static const int  kWindowHeight    = 513;

/* A notification code communicated from the load thread to the main application so that
 * it can display meaningful information about how the program is operating.
 */
static const UINT kProgressMessage = WM_APP + 1; 

/* An enumerated type listing possible progress messages. */
enum { Loading, Done, Failure };

/* A notification code used by the image to communicate to the main window that it's
 * been clicked on.
 */
static const size_t kNotifyCode   = 137;

/* Used in inverting user clicks to longitude and latitude space. */
static const double kPi = 3.1415926535897932364626433;

static bool InitializeWindow(HINSTANCE hInst, HWND hWnd);

/* Conversion functions to convert between C++ strings and TCHAR strings. */
static tstring WidenString(string str) {
	tstring result;
	for (size_t i = 0; i < str.length(); ++i)
		result += str[i];
	return result;
}

/* Given a window, retrieves the auxiliary data associated with that window. */
static WindowAuxiliaryData* GetAuxData(HWND hWnd) {
	return reinterpret_cast<WindowAuxiliaryData*>(GetWindowLongPtr(hWnd, GWLP_USERDATA));
}

/* Gets the cursor position relative to some window. */
static Point<2> GetClickLocation(HWND relativeTo) {
	RECT rc;
	POINT pt = {0, 0};

	/* Get the absolute cursor position and absolute window position. 
	 * If unable to do so, hand back the origin as a sentinel.
	 */
	if (!GetWindowRect(relativeTo, &rc) || !GetCursorPos(&pt)) {
		Point<2> origin;
		origin[0] = origin[1] = 0.0;
		return origin;
	}

	/* Adjust for absolute window position. */
	pt.x -= rc.left;
	pt.y -= rc.top;

	/* Normalize everything into [-1, 1] x [-1, 1] */
	Point<2> result;
	result[0] = (pt.x - kWindowWidth / 2)  / double(kWindowWidth / 2);
	result[1] = (pt.y - kWindowHeight / 2) / double(kWindowHeight / 2);

	return result;
}

/* This code is based on math from Wolfram MathWorld at
 *
 * http://mathworld.wolfram.com/MollweideProjection.html
 *
 * It inverts the click from a point in the window into
 * a longitude and latitude.
 */
static void InvertMollwideProjection(Point<2>& location) {
	/* Compute theta = arcsin(y) */
	double theta = asin(location[1]);

	/* Longitude is pi x / cos(theta) */
	double longitude = kPi * location[0] / cos(theta);
	
	/* Latitude is arcsin((2theta + sin(2theta)) / pi).  However,
	 * in our coordinate system y grows down instead of up, and so
	 * we need to negate the result.
	 */
	double latitude = -asin((2.0 * theta + sin(2.0 * theta)) / kPi);

	/* Convert from rads to degrees. */
	longitude *= 180.0 / kPi;
	latitude  *= 180.0 / kPi;

	/* Update the vector. */
	location[0] = longitude;
	location[1] = latitude;
}

/* Given a FIPS 10-4 name (i.e. US04), return the display name of that location
 * (i.e Arizona, United States).
 */
static string NameForLocation(const string& name, const map<string, string>& geoCodes) {
	/* Case one: We know what this is. */
	map<string, string>::const_iterator itr = geoCodes.find(name);
	if (itr != geoCodes.end()) return itr->second;

	/* Case two: We don't know what this is, but we know the country code. */
	itr = geoCodes.find(name.substr(0, 2));
	if (itr != geoCodes.end()) return itr->second;

	/* Case three: No idea.  Say so. */
	return "Unknown Location (FIPS 10-4 code: " + name + ")";
}

/* When the map is clicked on, convert the click to a longitude/latitude and 
 * process.
 */
static void ProcessClick(HWND hWnd, HWND hImageWnd) {
	/* If the data structures are not yet initialized, don't execute any queries */
	WindowAuxiliaryData* mapLookup = GetAuxData(hWnd);
	if (!mapLookup || !mapLookup->isReady || !mapLookup->hStatus)
		return;

	/* Figure out where the user clicked and give up if the click location
	 * is out of the unit circle.
	 */
	Point<2> location = GetClickLocation(hImageWnd);
	if (inner_product(location.begin(), location.end(), location.begin(), 0.0) >= 1)
		return;

	/* Convert from a point in the unit circle to a longitude and latitude. */
	InvertMollwideProjection(location);

	/* Report the nearest neighbor. */
	SetWindowText(mapLookup->hStatus,
				  WidenString(NameForLocation(mapLookup->kdt.kNNValue(location, 1), 
										      mapLookup->geoLookup)).c_str());
}

/* Respond to a status update about how the program is loading. */
static void ReportLoadStatus(HWND hWnd, WPARAM wParam, LPARAM lParam) {
	/* Get auxiliary data and abort if the status window hasn't been created yet. */
	WindowAuxiliaryData* aux = GetAuxData(hWnd);
	if (!aux || !aux->hStatus)
		return;

	/* Based on the status code, perform different operations. */
	switch (lParam) {
	/* Loading progress indicator. */
	case Loading:
		{
			stringstream converter;
			converter << "MapLookup: Loaded " << wParam << " locations...";
			SetWindowText(aux->hStatus, WidenString(converter.str()).c_str());
			break;
		}

	/* Indexing complete. */
	case Done:
		SetWindowText(aux->hStatus, _T("Ready"));
		break;

	/* Something went wrong: kill the program. */
	case Failure:
		MessageBox(hWnd, _T("An error occurred loading data.  MapLookup will now terminate"), _T("Fatal Error"), MB_ICONSTOP | MB_OK);
		DestroyWindow(hWnd);
		break;
	}
}

/* Main window's event handler.  This function is implicitly called by the system whenever
 * an event occurs.  We use this to dispatch commands to the proper handlers.
 */
LRESULT CALLBACK HandleMessage (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	/* If the window closes, quit the program. */
	if (message == WM_DESTROY) {
		PostQuitMessage(0);
		return 0;
	}
	
	/* If the user clicked on the map, respond to the click. */
	else if (message == WM_COMMAND && LOWORD(wParam) == kNotifyCode && HIWORD(wParam) == STN_CLICKED) {
		ProcessClick(hWnd, reinterpret_cast<HWND>(lParam));
		return 0;
	}

	/* If this is an internal progress message, take appropriate steps. */
	else if (message == kProgressMessage) {
		ReportLoadStatus(hWnd, wParam, lParam);
		return 0;
	}
	
	/* Otherwise just do whatever the system default is. */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/* Configures the window class for the main window.  This is all fairly technical
 * and is completely uninteresting.
 */
static bool InitializeWindowClass(HINSTANCE hInst) {
	WNDCLASSEX windowClass;
	windowClass.cbSize = sizeof(WNDCLASSEX);
	windowClass.style = CS_OWNDC | CS_DBLCLKS;
	windowClass.lpfnWndProc = HandleMessage;
	windowClass.cbWndExtra = 0;
	windowClass.cbClsExtra = 0;
	windowClass.hInstance = hInst;

	windowClass.hIcon = LoadIcon(hInst, _T("icon"));
	windowClass.hCursor = LoadCursor(NULL, IDC_CROSS);
	windowClass.hbrBackground = (HBRUSH)COLOR_APPWORKSPACE;
	windowClass.lpszMenuName = NULL;
	windowClass.lpszClassName = kWindowClassName;
	windowClass.hIconSm = LoadIcon(hInst, _T("icon"));
	
	/* Make sure registration worked! */
	return RegisterClassEx(&windowClass) != 0;
}

/* Loads the list of FIPS 10-4 codes from the auxiliary file. */
static bool LoadGeoCodes(map<string, string>& geoLookup) {
	/* Load geopolitical codes. */
	ifstream geoCodes("geo-codes.txt", ios::binary);
	if (!geoCodes) return false;

	string line;
	while (getline(geoCodes, line)) {
		/* Split the line into the code and the name of the place. */
		stringstream parser(line);
		string code, name;
		parser >> code;     // Get code
		parser.ignore(1);   // Skip tab
		getline(parser, name, '\r'); // Get name

		/* If this is a two-letter designation, it's a country name.  Store it. */
		if (code.length() == 2)
			geoLookup[code] = name;

		/* If this is a four-letter designation, it's the name of a state/province/etc.
		 * Set the name to be the name of the location suffixed with the name of the
		 * country.
		 */
		else if (code.length() == 4) // Dependent name
			geoLookup[code] = name + ", " + geoLookup[code.substr(0, 2)];
		
		/* Fail otherwise. */
		else
			return false;
	}
	return true;
}

/* Loads all of the locations and their designations. */
static bool LoadGeographicData(KDTree<2, string>& kd, HWND hWnd) {
	/* Load geographic data. */
	ifstream input("place-data.txt");
	if (!input) return false;

	/* Figure out how many there are and preallocate space for them. */
	size_t totalNumber;
	if(!(input >> totalNumber)) return false;

	/* Load all data. */
	int count = 0;
	string name;
	Point<2> pt;
	while (input >> pt[0] >> pt[1] >> name) {
		kd.insert(pt, name);
		if (++count % 10000 == 0)
			PostMessage(hWnd, kProgressMessage, count, Loading);
	}

	/* Confirm that we read enough data from the file. */
	return count == totalNumber;
}

/* Thread to load map data while the program starts up. */
static DWORD WINAPI LoadLocationData(LPVOID data) {
	/* Get the window handle. */
	HWND hWnd = static_cast<HWND>(data);
	WindowAuxiliaryData* lookupTable = GetAuxData(hWnd);

	/* Make sure auxiliary data is available before loading! */
	if (lookupTable == NULL) {
		PostMessage(hWnd, kProgressMessage, 0, Failure);
		return 0;
	}

	/* Load in the mapping from FIPS 10-4 codes to actual places. */
	if (!LoadGeoCodes(lookupTable->geoLookup)) {
		PostMessage(hWnd, kProgressMessage, 0, Failure);
		return 0;
	}

	/* Load the geographic data */
	if (!LoadGeographicData(lookupTable->kdt, hWnd)) {
		PostMessage(hWnd, kProgressMessage, 0, Failure);
		return 0;
	}

	/* Mark that we've loaded. */
	lookupTable->isReady = true;

	/* Report success! */
	PostMessage(hWnd, kProgressMessage, 0, Done);
	return 1;
}

/* Fires off a thread to load in the data.  This makes the program responsive even when
 * it's doing massive computations behind-the-scenes.
 */
static bool InitializeLoadingThread(HWND hWnd) {
	return CreateThread(NULL, 0, LoadLocationData, hWnd, 0, NULL) != NULL;
}

static bool InitializeWindow(HINSTANCE hInst, HWND hWnd) {
	/* Create the map display. */
	if (CreateWindowEx(0, _T("static"), _T("world-map"), SS_NOTIFY | WS_CHILD | WS_VISIBLE | SS_BITMAP, 
					   0, 0, 0, 0, 
					   hWnd, reinterpret_cast<HMENU>(kNotifyCode),
					   hInst, NULL) == NULL) 
					   return false;

	/* Force the window to be the right size. */
	RECT rv;
	GetWindowRect(hWnd, &rv);
	POINT windowPos = {rv.left, rv.top};

	/* Get a rectangle holding the new coordinates */
	RECT rc;
	SetRect(&rc, 0, 0, kWindowWidth, kWindowHeight);
	AdjustWindowRectEx(&rc, WS_BORDER | WS_SYSMENU, FALSE, 0);
	
	/* Put the window there */
	MoveWindow(hWnd, rc.left, rc.top, rc.right-rc.left, rc.bottom-rc.top, TRUE);

	/* Get new size */
	RECT newsize;
	GetClientRect(hWnd, &newsize);

	/* Adjust based on new size information */
	rc.right  += kWindowWidth - newsize.right;
	rc.bottom += kWindowHeight - newsize.bottom;

	/* Shift */
	OffsetRect(&rc, windowPos.x, windowPos.y);

	/* Move and show window */
	MoveWindow(hWnd, rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top, TRUE);

	WindowAuxiliaryData* lookupTable = new WindowAuxiliaryData;
	lookupTable->isReady = false;

	/* Create the status control. */
	lookupTable->hStatus = CreateWindowEx(0, STATUSCLASSNAME, _T("Loading..."), WS_CHILD | WS_VISIBLE,
		                                  0, 0, 0, 0, 
										  hWnd, reinterpret_cast<HMENU>(kNotifyCode + 1), hInst, NULL);
	if (lookupTable->hStatus == NULL)
		return false;
	
	/* Update the window aux data so that we can look this up later. */
	SetLastError(0);
	if(!SetWindowLongPtr(hWnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(lookupTable)) && GetLastError())
		return false;

	return true;
}

/* Main program entry point.  Set up the window and wait for user input. */
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, int) {
	InitCommonControls();
	if (!InitializeWindowClass(hInst)) 
		return 0;
	
	/* Create (but don't show) the window */
	HWND mainWindow = CreateWindowEx(0, kWindowClassName, _T("MapLookup"), WS_VISIBLE | WS_BORDER | WS_SYSMENU, 
		                             (GetSystemMetrics(SM_CXSCREEN) - kWindowWidth) / 2,
									 (GetSystemMetrics(SM_CYSCREEN) - kWindowHeight) / 2 , 
									 kWindowWidth, kWindowHeight, 
									 NULL, NULL, hInst, NULL);
	
	/* Setup the window; fail if unable to do so. */
	if(!mainWindow || !InitializeWindow(hInst, mainWindow))
		return 0;

	/* Start up a thread to load in all of the program data. */
	if (!InitializeLoadingThread(mainWindow))
		return 0;

	/* Main program loop.  Wait for a message,then go dispatch it. */
	while(true) {
		WaitMessage();

		MSG message;
		while(PeekMessage(&message, NULL, 0, 0, PM_REMOVE)) {			
			/* On exit, kill the program. */
			if(message.message == WM_QUIT)
				return int(message.wParam);
			
			TranslateMessage(&message);
			DispatchMessage(&message);
		}
	}
}
